import argparse
from pathlib import Path

from qat_rpc.zmq.wrappers import ZMQClient

parser = argparse.ArgumentParser(
    prog="QAT submission service", description="Submit your QASM or QIR program to QAT."
)
parser.add_argument("program", type=str, help="Program string or path to program file.")
parser.add_argument("--config", type=str, help="Serialised CompilerConfig json")


def qat_run():
    args = parser.parse_args()
    program = args.program
    config = args.config
    if Path(program).is_file():
        program = Path(program).read_text()
    if config is not None and Path(config).is_file():
        config = Path(config).read_text()
    zmq_client = ZMQClient()
    results = zmq_client.execute_task(program, config)
    print(results)



from enum import Enum
from importlib.metadata import version
from time import time
from typing import Optional, Union

import zmq
from compiler_config.config import CompilerConfig
from qat.purr.backends.echo import get_default_echo_hardware
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.runtime import get_runtime
from qat.purr.integrations.features import OpenPulseFeatures
from qat.qat import execute_with_metrics

from qat_rpc.utils.metrics import MetricExporter


class Messages(Enum):
    PROGRAM = "program"
    VERSION = "version"
    COUPLINGS = "couplings"
    QUBIT_INFO = "qubit_info"
    QPU_INFO = "qpu_info"


class ZMQBase:
    def __init__(
        self, socket_type: zmq.SocketType, ip_address: str = "127.0.0.1", port: str = "5556"
    ):
        self._context = zmq.Context()
        self._socket = self._context.socket(socket_type)
        self._timeout = 30.0
        self._protocol = "tcp"
        self._ip_address = ip_address
        self._port = port

    @property
    def address(self):
        return f"{self._protocol}://{self._ip_address}:{self._port}"

    def _check_recieved(self):
        try:
            msg = self._socket.recv_pyobj(zmq.NOBLOCK)
            return msg
        except zmq.ZMQError:
            return None

    def _send(self, message) -> None:
        sent = False
        t0 = time()
        while not sent:
            try:
                self._socket.send_pyobj(message, zmq.NOBLOCK)
                sent = True
            except zmq.ZMQError:
                if time() > t0 + self._timeout:
                    raise TimeoutError(
                        "Sending %s on %s timedout" % (message, self.address)
                    )
        return

    def close(self):
        """Disconnect the link to the socket."""
        if self._socket.closed:
            return
        self._socket.close()
        self._context.destroy()

    def __del__(self):
        self.close()


class ZMQServer(ZMQBase):
    def __init__(
        self,
        hardware: Optional[QuantumHardwareModel] = None,
        server_port: Optional[str] = "5556",
        metric_exporter: Optional[MetricExporter] = None,
    ):
        super().__init__(socket_type=zmq.REP, port=server_port)
        self._metric = metric_exporter
        self._socket.bind(self.address)
        self._hardware = hardware or get_default_echo_hardware(qubit_count=32)
        self._engine = get_runtime(self._hardware).engine
        self._running = False

    @property
    def address(self):
        return f"{self._protocol}://*:{self._port}"

    def _program(self, program, config):
        program = program
        config = CompilerConfig.create_from_json(config)
        result, metrics = execute_with_metrics(program, self._engine, config)
        return {"results": result, "execution_metrics": metrics}

    def _version(self):
        return {"qat_rpc_version": str(version("qat_rpc"))}

    def _couplings(self):
        coupling = [
            coupled.direction for coupled in self._hardware.qubit_direction_couplings
        ]
        return {"couplings": coupling}

    def _qubit_info(self):
        raise NotImplementedError(
            "Individual qubit information not implented, pending hardware model changes."
        )

    def _qpu_info(self):
        features = OpenPulseFeatures()
        features.for_hardware(self._hardware)
        qpu_info = features.to_json_dict()
        return {"qpu_info": qpu_info}

    def _interpret_message(self, message):
        match message[0]:
            case Messages.PROGRAM.value:
                print(message)
                if len(message) != 3:
                    raise ValueError(
                        f"Program message should be of length 3, not {len(message)}"
                    )
                return self._program(message[1], message[2])
            case Messages.VERSION.value:
                return self._version()
            case Messages.COUPLINGS.value:
                return self._couplings()
            case Messages.QUBIT_INFO:
                return self._qubit_info()
            case Messages.QPU_INFO.value:
                return self._qpu_info()

            case _:
                return self._program(message[0], message[1])

    def run(self):
        self._running = True
        with self._metric.receiver_status() as metric:
            metric.succeed()
        while self._running and not self._socket.closed:
            msg = self._check_recieved()
            if msg is not None:
                try:
                    reply = self._interpret_message(message=msg)
                    with self._metric.executed_messages() as executed:
                        executed.increment()
                except Exception as e:
                    reply = {"Exception": repr(e)}
                    with self._metric.failed_messages() as failed:
                        failed.increment()
                self._send(reply)

    def stop(self):
        self._running = False
        with self._metric.receiver_status() as metric:
            metric.fail()


class ZMQClient(ZMQBase):
    def __init__(
        self, client_ip: Optional[str] = "127.0.0.1", client_port: Optional[str] = "5556"
    ):
        super().__init__(socket_type=zmq.REQ, ip_address=client_ip, port=client_port)
        self._socket.setsockopt(zmq.LINGER, 0)
        self._socket.connect(self.address)

    def _send(self, message):
        super()._send(message=message)
        return self._await_results()

    def _await_results(self):
        result = None
        while result is None:
            result = self._check_recieved()
        return result

    def execute_task(self, program: str, config: Union[CompilerConfig, str] = None):
        self.result = None
        if isinstance(config, str):
            # Verify config string is valid before submitting.
            config = CompilerConfig.create_from_json(config)
        cfg = config or CompilerConfig()
        return self._send((Messages.PROGRAM.value, program, cfg.to_json()))

    def api_version(self):
        return self._send((Messages.VERSION.value,))

    def qpu_couplings(self):
        return self._send((Messages.COUPLINGS.value,))

    def qubit_info(self):
        return self._send((Messages.QUBIT_INFO.value,))

    def qpu_info(self):
        return self._send((Messages.QPU_INFO.value,))



import abc
import inspect
import re
from inspect import getmembers, ismethod
from typing import Callable, Union

from prometheus_client import Counter, Gauge, start_http_server
from qat.purr.utils.logger import get_default_logger

from qat_rpc.utils.constants import PROMETHEUS_PORT

log = get_default_logger()


class IncrementMutableOutcome:
    def __init__(self):
        self._count: float = 0.0

    def increment(self, amount: float = 1.0):
        self._count += amount

    def __float__(self):
        return self._count

    def __int__(self):
        return int(self._count)


class BinaryMutableOutcome:
    def __init__(self):
        self._success: bool = None

    def succeed(self):
        # Failure is sticky
        if self._success is None or self._success:
            self._success = True

    def fail(self):
        self._success = False

    def __int__(self):
        return int(self._success)

    def __float__(self):
        return 1.0 if self._success else 0.0

    def __eq__(self, other):
        if isinstance(other, BinaryMutableOutcome):
            return other is not None and other._success == self._success
        if isinstance(other, bool):
            return (self is not None and not other) or (other == self._success)


class ReceiverBackend(abc.ABC):
    @abc.abstractmethod
    def receiver_status(self, outcome: BinaryMutableOutcome): ...

    @abc.abstractmethod
    def failed_messages(self, outcome: IncrementMutableOutcome): ...

    @abc.abstractmethod
    def executed_messages(self, outcome: IncrementMutableOutcome): ...

    @abc.abstractmethod
    def hardware_connected(self, outcome: BinaryMutableOutcome): ...

    @abc.abstractmethod
    def hardware_reloaded(self, outcome: BinaryMutableOutcome): ...


class NullReceiverBackend(ReceiverBackend):
    def __init__(self):
        super(NullReceiverBackend, self).__init__()

    def receiver_status(self, outcome: BinaryMutableOutcome): ...

    def failed_messages(self, outcome: IncrementMutableOutcome): ...

    def executed_messages(self, outcome: IncrementMutableOutcome): ...

    def hardware_connected(self, outcome: BinaryMutableOutcome): ...

    def hardware_reloaded(self, outcome: BinaryMutableOutcome): ...


class PrometheusReceiver(ReceiverBackend):
    def __init__(self, port: int = PROMETHEUS_PORT):
        super(PrometheusReceiver, self).__init__()
        start_http_server(port)
        log.info(f"Starting Prometheus metrics exporter on port {port}.")

        self._receiver_status = Gauge(
            "receiver_status", "Measure the Receiver backend up state"
        )
        self._failed_messages = Counter("failed_messages", "messages failure counter")
        self._executed_messages = Counter("executed_messages", "messages executed counter")
        self._hardware_connected_status = Gauge(
            "hardware_connected_status", "Indicate connected status of live hardware"
        )
        self._hardware_reloaded_status = Gauge(
            "hardware_reloaded_status",
            "Indicate if hardware reload from calibration succeeded or failed",
        )

    def receiver_status(self, outcome: BinaryMutableOutcome):
        self._receiver_status.set(outcome)

    def failed_messages(self, outcome: IncrementMutableOutcome):
        self._failed_messages.inc(float(outcome))

    def executed_messages(self, outcome: IncrementMutableOutcome):
        self._executed_messages.inc(float(outcome))

    def hardware_connected(self, outcome: BinaryMutableOutcome):
        self._hardware_connected_status.set(outcome)

    def hardware_reloaded(self, outcome: BinaryMutableOutcome):
        self._hardware_reloaded_status.set(outcome)


class ReceiverAdapter(ReceiverBackend):
    """
    Presents a normal derived type to the MetricsExporter which is required for
    introspection as part of it's dynamic build. When using Mocks, which replace
    functions with variables, it is required to first wrap those in Adapters.
    """

    def __init__(self, decorated):
        self.decorated = decorated

    def receiver_status(self, outcome: BinaryMutableOutcome):
        self.decorated.receiver_status(outcome)

    def failed_messages(self, outcome: IncrementMutableOutcome):
        self.decorated.failed_messages(outcome)

    def executed_messages(self, outcome: IncrementMutableOutcome):
        self.decorated.executed_messages(outcome)

    def hardware_connected(self, outcome: BinaryMutableOutcome):
        self.decorated.hardware_connected(outcome)

    def hardware_reloaded(self, outcome: BinaryMutableOutcome):
        self.decorated.hardware_reloaded(outcome)


class MetricFieldWrapper:
    def __init__(
        self,
        func: Callable,
        outcome: Union[IncrementMutableOutcome, BinaryMutableOutcome],
    ):
        self.callable = func
        self.outcome = outcome

    def __enter__(self):
        # Loan out the mutable outcome for change
        return self.outcome

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Apply the outcome to target callable
        try:
            self.callable(self.outcome)
        except Exception as ex:
            log.warning(f"Metric setting errored {str(ex)}")


class MetricExporter:
    """
    Factory for context managers for metrics reporting over different
    backend reporting frameworks

    Dynamically builds member functions that mirror the backend it is given.
    The Backend is the API specification.
    """

    def __init__(self, backend):
        if isinstance(backend, type):
            raise ValueError("Argument must be an instance not a type")
        methods = [
            member
            for member in getmembers(backend, predicate=ismethod)
            if not re.match("^_", member[0])
        ]

        def decorate(f):
            # factory function avoids closure issues
            try:
                sign = inspect.signature(f)
                parameters = sign.parameters
                outcome_param = parameters.get("outcome")
                outcome_type = outcome_param.annotation
                return lambda: MetricFieldWrapper(f, outcome_type())
            except KeyError as ex:
                log.error(
                    f" An error occurred while processing "
                    f"the function's outcome parameter {ex}"
                )
                raise ex

        for func_name, func in methods:
            nargs = func.__code__.co_argcount
            if not nargs == 2:
                raise ValueError(
                    f"Receiver must have one non-self argument on each "
                    f"function. Found {nargs} while evaluating {func_name}"
                    f" on {str(type(backend))}"
                )
            # build no-arg setting functions on this matching backend public name
            setattr(self, func_name, decorate(func))



import os
from pathlib import Path
from signal import SIGINT, SIGTERM, signal

from qat.purr.compiler.devices import Calibratable
from qat.purr.utils.logger import get_default_logger

from qat_rpc.utils.constants import PROMETHEUS_PORT
from qat_rpc.utils.metrics import MetricExporter, PrometheusReceiver
from qat_rpc.zmq.wrappers import ZMQServer

log = get_default_logger()


class GracefulKill:
    def __init__(self, receiver: ZMQServer):
        signal(SIGINT, self._sigint)
        signal(SIGTERM, self._sigterm)
        self.receiver = receiver

    def _sigint(self, *args):
        self.receiver.stop()

    def _sigterm(self, *args):
        self.receiver.stop()


def main():
    hw = None
    receiver_port = os.getenv("RECEIVER_PORT")
    metrics_port = os.getenv("METRICS_PORT")

    if receiver_port is not None:
        try:
            int(receiver_port)
        except ValueError:
            log.warning("Configured receiver port is not a valid integer.")
            receiver_port = 5556
            log.info(f"Defaulting receiver to run on port {receiver_port}")
        else:
            try:
                if not (1024 < int(receiver_port) < 49152):
                    raise ValueError(
                        "Receiver must be configured to run on a valid port number between 1024 and 49152."
                    )
                log.info(f"Receiver server is configured to start on port {receiver_port}.")
            except ValueError as e:
                log.warning(f"{e}")
                receiver_port = 5556
                log.info(f"Defaulting receiver to run on port {receiver_port}.")
    else:
        receiver_port = 5556

    if metrics_port is not None:
        try:
            int(metrics_port)
        except ValueError:
            log.warning("Configured metrics port is not a valid integer.")
            metrics_port = PROMETHEUS_PORT
            log.info(f"Defaulting metrics exporter to run on port {metrics_port}")
        else:
            try:
                if int(metrics_port) == int(receiver_port):
                    raise ValueError(
                        f"Metrics exporter cannot run on port {metrics_port}, it must be set to a different value to the receiver port."
                    )
                elif not (1024 < int(metrics_port) < 49152):
                    raise ValueError(
                        "Metrics exporter must be configured to run on a valid port number between 1024 and 49152."
                    )
                log.info(f"Metrics exporter is configured to start on port {metrics_port}.")
            except ValueError as e:
                log.warning(f"{e}")
                metrics_port = PROMETHEUS_PORT
                log.info(f"Defaulting metrics exporter to run on port {metrics_port}.")
    else:
        metrics_port = PROMETHEUS_PORT

    try:
        if int(metrics_port) == int(receiver_port):
            raise ValueError
    except ValueError:
        log.warning(
            "The receiver and metrics exporter have been configured to run on the same port."
        )
        receiver_port = 5556
        metrics_port = PROMETHEUS_PORT
        log.info(
            f"Defaulting receiver to run on port {receiver_port} and metrics exporter to run on port {metrics_port}."
        )

    metric_exporter = MetricExporter(backend=PrometheusReceiver(port=int(metrics_port)))

    if (calibration_file := os.getenv("TOSHIKO_CAL")) is not None:
        calibration_file = Path(calibration_file)
        if not calibration_file.is_absolute() and not calibration_file.is_file():
            calibration_file = Path(Path(__file__).parent, calibration_file)
        if not calibration_file.is_file():
            raise ValueError(f"No such file: {calibration_file}")
        log.info(f"Loading: {calibration_file} ")
        hw = Calibratable.load_calibration_from_file(str(calibration_file))
        log.debug("Loaded")

    receiver = ZMQServer(
        hardware=hw, server_port=receiver_port, metric_exporter=metric_exporter
    )
    gk = GracefulKill(receiver)

    log.info(
        f"Starting QAT receiver on port {str(receiver._port)} with {type(receiver._hardware)} hardware."
    )
    receiver.run()


if __name__ == "__main__":
    main()
